#include <vector>
#include <GL/glut.h>
#include "3ddatastruct.h"

extern Matrix4x4 _view_matrix;	// the active view matrix
extern Matrix4x4 _world_matrix;	// the active world matrix

//////////// vertex and triangle constructors ////////////////

Vertex::Vertex(const Vector3 &pos, const Color &col, float u, float v) {
	this->pos = pos;
	color = col;
	this->u = u;
	this->v = v;
}

Triangle::Triangle(Index v1, Index v2, Index v3) {
	verts[0] = v1;
	verts[1] = v2;
	verts[2] = v3;
}

/////////// mesh implementation ////////////////

Mesh::Mesh(unsigned long vc, unsigned long tc) {
	vertex_count = vc;
	triangle_count = tc;
	index_count = tc * 3;

	if(!vc || !tc) {
		varray = 0;
		tarray = 0;
		iarray = 0;
	} else {
		varray = new Vertex[vertex_count];
		tarray = new Triangle[triangle_count];
		iarray = new Index[index_count];
	}

	prs.pos = Vector3(0, 0, 0);
	prs.rot.ResetIdentity();
	prs.scale = Vector3(1, 1, 1);
}

Mesh::~Mesh() {
	if(varray) delete [] varray;
	if(tarray) delete [] tarray;
	if(iarray) delete [] iarray;
}

// SetGeometry() fills vertex and triangle data for our mesh (local copy)
// AND generates the index buffer from the triangle indices
void Mesh::SetGeometry(const Vertex *vbuf, const Triangle *tbuf, unsigned long vcount, unsigned long tcount) {
	if(vertex_count != vcount) {
		vertex_count = vcount;
		if(varray) delete [] varray;
		varray = new Vertex[vertex_count];
	}

	if(triangle_count != tcount) {
		triangle_count = tcount;
		if(tarray) delete [] tarray;
		tarray = new Triangle[triangle_count];

		index_count = tcount * 3;
		if(iarray) delete [] iarray;
		iarray = new Index[index_count];
	}

	memcpy(varray, vbuf, vertex_count * sizeof(Vertex));
	memcpy(tarray, tbuf, triangle_count * sizeof(Triangle));

	Index *iptr=iarray;
	for(unsigned int i=0; i<triangle_count; i++) {
		for(int j=0; j<3; j++) {
			*iptr++ = tarray[i].verts[j];
		}
	}

}

// accessors

Vertex * const Mesh::GetVertexArray() {
	return varray;
}

Triangle * const Mesh::GetTriangleArray() {
	return tarray;
}

Index * const Mesh::GetIndexArray() {
	return iarray;
}

unsigned long Mesh::GetVertexCount() const {
	return vertex_count;
}

unsigned long Mesh::GetTriangleCount() const {
	return triangle_count;
}

unsigned long Mesh::GetIndexCount() const {
	return index_count;
}

void Mesh::SetMaterial(const Material &mat) {
	this->mat = mat;
}

Material *Mesh::GetMaterial() {
	return &mat;
}

// functions to help with the manipulation of meshes in 3d space

void Mesh::ResetTransform() {
	prs.pos = Vector3(0, 0, 0);
	prs.rot.ResetIdentity();
	prs.scale = Vector3(1, 1, 1);
}

void Mesh::ResetTranslation() {
	prs.pos = Vector3(0, 0, 0);
}

void Mesh::ResetRotation() {
	prs.rot.ResetIdentity();
}

void Mesh::ResetScaling() {
	prs.scale = Vector3(1, 1, 1);
}

void Mesh::Translate(const Vector3 &tvec) {
	prs.pos += tvec;
}

void Mesh::Rotate(const Vector3 &axis, float angle) {
	prs.rot.Rotate(axis, angle);
}

void Mesh::Scale(const Vector3 &svec) {
	prs.scale.Scale(svec.x, svec.y, svec.z);
}

/////////////////////////////////////////

void CalculateTriangleNormals(Triangle *tarray, unsigned long tcount, const Vertex *varray) {

	for(unsigned long i=0; i<tcount; i++) {
		Vector3 v1 = varray[tarray[i].verts[0]].pos - varray[tarray[i].verts[1]].pos;
		Vector3 v2 = varray[tarray[i].verts[0]].pos - varray[tarray[i].verts[2]].pos;
		tarray[i].normal = -CrossProduct(v1, v2).Normalized();
	}
}

void CalculateVertexNormals(Vertex *varray, unsigned long vcount, const Triangle *tarray, unsigned long tcount) {

	// first calculate triangle - vertex adjacency list
	std::vector<Index> *adj = new std::vector<Index>[vcount];

	for(unsigned long i=0; i<tcount; i++) {
		for(int j=0; j<3; j++) {
            adj[tarray[i].verts[j]].push_back((Index)i);
		}
	}

	for(unsigned long i=0; i<vcount; i++) {

		unsigned int adjcount;
		if((adjcount = (unsigned long)adj[i].size()) < 1) continue;

		varray[i].normal = Vector3(0.0f, 0.0f, 0.0f);
		for(unsigned int j=0; j<adjcount; j++) {
			varray[i].normal += tarray[adj[i][j]].normal;
		}

		varray[i].normal /= (float)adjcount;
	}
}

// render the mesh
void Mesh::Render() {

	// setup material
	mat.specular = mat.specular_power > 0.0f ? Color(1.0f, 1.0f, 1.0f) : Color(0.0f, 0.0f, 0.0f);
	mat.SetGLMaterial();

	// set transformation
	Matrix4x4 modelview = _world_matrix * _view_matrix;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadMatrixf(ConvertMatrixGL(modelview));

	/*
	// prepare for rendering
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), varray);
	glColorPointer(4, GL_FLOAT, sizeof(Vertex), &varray->color);
	glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), &varray->u);
	glNormalPointer(GL_FLOAT, sizeof(Vertex), &varray->normal);
	*/

	glEnable(GL_TEXTURE_2D);
	if(mat.texmap) {
		glBindTexture(GL_TEXTURE_2D, mat.texmap);
	}

	/*
	// render the data
	glDrawElements(GL_TRIANGLES, triangle_count * 3, GL_UNSIGNED_SHORT, iarray);
	*/

	Index *iptr = iarray;
	glBegin(GL_TRIANGLES);
	for(int i=0; i<triangle_count; i++) {
		for(int j=0; j<3; j++) {
			Vertex *vptr = varray + *iptr++;
			glColor4f(vptr->color.r, vptr->color.g, vptr->color.b, vptr->color.a);
			glNormal3f(vptr->normal.x, vptr->normal.y, vptr->normal.z);
			glTexCoord2f(vptr->u, vptr->v);
			glVertex3f(vptr->pos.x, vptr->pos.y, vptr->pos.z);
		}
	}
	glEnd();
	

	glDisable(GL_TEXTURE_2D);

	/*
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	*/

	

	glPopMatrix();
}

/////// light implementation ///////

Light::Light(const Vector3 &pos, const Color &col, float intensity) {
	this->pos = pos;
	this->col = col;
	this->intensity = intensity;
	attenuate = false;
}

void Light::SetPosition(const Vector3 &pos) {
	this->pos = pos;
}

Vector3 Light::GetPosition() const {
	return pos;
}

void Light::SetIntensity(float intensity) {
	this->intensity = intensity;
}

float Light::GetIntensity() const {
	return intensity;
}

void Light::SetAttenuation(bool enable) {
	attenuate = enable;
}

void Light::SetLight(int n) {
	int light_num = GL_LIGHT0 + n;

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadMatrixf(ConvertMatrixGL(_view_matrix));

	Color col = this->col * intensity;

	glLightfv(light_num, GL_POSITION, (float*)&pos);
	glLightfv(light_num, GL_DIFFUSE, (float*)&col);
	glLightfv(light_num, GL_SPECULAR, (float*)&col);

	if(attenuate) {
        glLightf(light_num, GL_CONSTANT_ATTENUATION, 1.0f);
		glLightf(light_num, GL_LINEAR_ATTENUATION, 0.0f);
		glLightf(light_num, GL_QUADRATIC_ATTENUATION, 0.006f);
	}

	glEnable(light_num);

	glPopMatrix();
}

/////////////// camera implementation /////////////

Camera::Camera() {}

void Camera::ResetTransform() {
	prs.pos = Vector3(0, 0, 0);
	prs.rot.ResetIdentity();
	prs.scale = Vector3(1, 1, 1);
}

void Camera::ResetTranslation() {
	prs.pos = Vector3(0, 0, 0);
}

void Camera::ResetRotation() {
	prs.rot.ResetIdentity();
}

void Camera::Translate(const Vector3 &tvec) {
	prs.pos += tvec;
}

void Camera::Rotate(const Vector3 &axis, float angle) {
	prs.rot.Rotate(axis, angle);
}

void Camera::UpdateViewMatrix() {
	Matrix4x4 trans, rot;

	trans.Translate(-prs.pos.x, -prs.pos.y, -prs.pos.z);
	rot = (Matrix4x4)prs.rot.Inverse().GetRotationMatrix();

	_view_matrix = trans * rot;
}
