#ifndef _3DDATASTRUCT_H_
#define _3DDATASTRUCT_H_

// in this file there are various data structures needed for 3D data

#include "3dmath.h"
#include "color.h"
#include "material.h"

typedef unsigned short Index;
class Vertex;
class Triangle;

// usefull functions for normal vector calculations (for lighting)
void CalculateTriangleNormals(Triangle *tarray, unsigned long tcount, const Vertex *varray);
void CalculateVertexNormals(Vertex *varray, unsigned long vcount, const Triangle *tarray, unsigned long tcount);

// vertex class: holds position, normal (for lighting), color, and texture coordinates
class Vertex {
public:
	Vector3 pos;
	Vector3 normal;
	Color color;
	float u, v;

	Vertex(const Vector3 &pos=Vector3(0,0,0), const Color &col=1.0f, float u=0.0f, float v=0.0f);
};

// triangle class: with face normal and 3 indices into a vertex array
class Triangle {
public:
	Index verts[3];
	Vector3 normal;

	Triangle(Index v1=0, Index v2=0, Index v3=0);
};

// Position/Rotation/Scale (a convinient structure for manipulation of objects in 3D space)
// position and scaling are represented by vectors, orientation by a quaternion
struct PRS {
	Vector3 pos;
	Quaternion rot;
	Vector3 scale;
};

// Mesh class: holds vertex and triangle arrays, as well as material properties
class Mesh {
private:
	Vertex *varray;
	Triangle *tarray;
	Index *iarray;
	unsigned long vertex_count, triangle_count, index_count;

	Material mat;

	PRS prs;

public:
	Mesh(unsigned long vc=0, unsigned long tc=0);
	~Mesh();

	void SetGeometry(const Vertex *vbuf, const Triangle *tbuf, unsigned long vcount, unsigned long tcount);
	
	Vertex * const GetVertexArray();
	Triangle * const GetTriangleArray();
	Index * const GetIndexArray();

	unsigned long GetVertexCount() const;
	unsigned long GetTriangleCount() const;
	unsigned long GetIndexCount() const;

	void SetMaterial(const Material &mat);
	Material *GetMaterial();

	void ResetTransform();
	void ResetTranslation();
	void ResetRotation();
	void ResetScaling();

	void Translate(const Vector3 &tvec);
	void Rotate(const Vector3 &axis, float angle);
	void Scale(const Vector3 &svec);

	void Render();
};

// a class for the light sources, call SetLight() when making changes to update
// the OpenGL states
class Light {
private:
	Vector3 pos;
	Color col;
	float intensity;
	bool attenuate;

public:
	Light(const Vector3 &pos=Vector3(0.0f, 0.0f, 0.0f), const Color &col=Color(1.0f, 1.0f, 1.0f), float intensity=1.0f);

	void SetPosition(const Vector3 &pos);
	Vector3 GetPosition() const;

	void SetIntensity(float intensity);
	float GetIntensity() const;

	void SetAttenuation(bool enable);

	void SetLight(int n);
};


// a class to abstract the concept of a camera
class Camera {
private:
	PRS prs;

public:
	Camera();

	void ResetTransform();
	void ResetTranslation();
	void ResetRotation();

	void Translate(const Vector3 &tvec);
	void Rotate(const Vector3 &axis, float angle);

	inline Vector3 GetPosition() {return prs.pos;}
	inline Quaternion GetRotation() {return prs.rot;}

	void UpdateViewMatrix();
};


#endif	// _3DDATASTRUCT_H_
