#include <cmath>
#include "color.h"

typedef uint32_t dword;
typedef uint16_t word;
typedef uint8_t byte;

template <class T>
const T &clamp(const T &val, const T &low, const T &high) {
	if(val > high) return high;
	if(val < low) return low;
	return val;
}

Color::Color(float intensity) {
	r = g = b = clamp<float>(intensity, 0.0f, 1.0f);
	a = 1.0f;
}

Color::Color(float r, float g, float b, float a) {
	this->r = clamp<float>(r, 0.0f, 1.0f);
	this->g = clamp<float>(g, 0.0f, 1.0f);
	this->b = clamp<float>(b, 0.0f, 1.0f);
	this->a = clamp<float>(a, 0.0f, 1.0f);
}

Color::Color(unsigned int r, unsigned int g, unsigned int b, unsigned int a) {
	this->r = (float)clamp<unsigned int>(r, 0, 255) / 255.0f;
	this->g = (float)clamp<unsigned int>(g, 0, 255) / 255.0f;
	this->b = (float)clamp<unsigned int>(b, 0, 255) / 255.0f;
	this->a = (float)clamp<unsigned int>(a, 0, 255) / 255.0f;
}

Color::Color(int r, int g, int b, int a) {
	this->r = (float)clamp<int>(r, 0, 255) / 255.0f;
	this->g = (float)clamp<int>(g, 0, 255) / 255.0f;
	this->b = (float)clamp<int>(b, 0, 255) / 255.0f;
	this->a = (float)clamp<int>(a, 0, 255) / 255.0f;
}

Color::Color(long pcol) {
	a = (float)(((dword)pcol >> 24) & 0xff) / 255.0f;
	r = (float)(((dword)pcol >> 16) & 0xff) / 255.0f;
	g = (float)(((dword)pcol >> 8) & 0xff) / 255.0f;
	b = (float)((dword)pcol & 0xff) / 255.0f;
}

Color::Color(dword pcol) {
	a = (float)((pcol >> 24) & 0xff) / 255.0f;
	r = (float)((pcol >> 16) & 0xff) / 255.0f;
	g = (float)((pcol >> 8) & 0xff) / 255.0f;
	b = (float)(pcol & 0xff) / 255.0f;
}

dword Color::GetPacked32() const {
	return	(((dword)(a * 255.0f) << 24) & 0xff000000) | 
			(((dword)(r * 255.0f) << 16) & 0x00ff0000) | 
			(((dword)(g * 255.0f) << 8) & 0x0000ff00) | 
			((dword)(b * 255.0f) & 0x000000ff);
}

word Color::GetPacked16() const {
	return (word)(r * 32.0f) << 11 | (word)(g * 64.0f) << 5 | (word)(b * 32.0f);
}

word Color::GetPacked15() const {
	return (word)a << 15 | (word)(r * 32.0f) << 10 | (word)(g * 32.0f) << 5 | (word)(b * 32.0f);
}

static const float HalfPi = 1.5707963268f;

byte Color::GetNearest8(const byte **pal) const {
	float Score[256];
	for(int i=0; i<256; i++) {
		Color palcol = Color(pal[i][0], pal[i][1], pal[i][2]);
		float NearR = (float)cos(fabs(r - palcol.r) * HalfPi);
		float NearG = (float)cos(fabs(g - palcol.g) * HalfPi);
		float NearB = (float)cos(fabs(b - palcol.b) * HalfPi);
		Score[i] = NearR + NearG + NearB;
	}

	int nearest = 0;
	for(int i=0; i<256; i++) {
		if(Score[i] > Score[nearest]) nearest = i;
	}
	return nearest;
}

Color Color::operator +(const Color &col) const {
	return Color(r + col.r, g + col.g, b + col.b, a + col.a);
}

Color Color::operator -(const Color &col) const {
	return Color(r - col.r, g - col.g, b - col.b, a - col.a);
}

Color Color::operator *(const Color &col) const {
	return Color(r * col.r, g * col.g, b * col.b, a * col.a);
}
	
Color Color::operator *(float scalar) const {
	return Color(r * scalar, g * scalar, b * scalar, a);
}


void Color::operator +=(const Color &col) {
	*this = Color(r + col.r, g + col.g, b + col.b, a + col.a);
}

void Color::operator -=(const Color &col) {
	*this = Color(r - col.r, g - col.g, b - col.b, a - col.a);
}

void Color::operator *=(const Color &col) {
	*this = Color(r * col.r, g * col.g, b * col.b, a * col.a);
}

void Color::operator *=(float scalar) {
	*this = Color(r * scalar, g * scalar, b * scalar, a);
}

Color BlendColors(const Color &c1, const Color &c2, float t) {
	float r = c1.r + (c2.r - c1.r) * t;
	float g = c1.g + (c2.g - c1.g) * t;
	float b = c1.b + (c2.b - c1.b) * t;
	return Color(r, g, b);
}
