#include <iostream>
#include <fstream>
#include <unistd.h>
#include <GL/glut.h>
#include "terrain.h"
#include "timer.h"

using std::cout;
using std::cerr;
using std::endl;

bool Init();
void MainLoop();
void CleanUp();
void KeyHandler(unsigned char key, int x, int y);
void SpecialKeyHandler(int key, int x, int y);
void SpecialKeyUpHandler(int key, int x, int y);
void motion_handler(int x, int y);
void ProcessKeyPresses();

int ScreenX = 1024;
int ScreenY = 768;

int win;
Level lvl;
Camera cam;	// REMOVE when entities are done, replace with player
Light light;
bool fullscreen = false;
ntimer timer;
bool running = false;
bool mouse_enabled = false;

int main(int argc, char **argv) {

	glutInit(&argc, argv);

	for(int i=1; i<argc; i++) {
		if(!strcmp(argv[i], "-size")) {
			if(argv[++i] && isdigit(argv[i][0])) {
				ScreenX = atoi(argv[i]);
				char *ptr = strchr(argv[i], 'x');
				if(ptr && isdigit(*++ptr)) {
					ScreenY = atoi(ptr);
				} else {
					cerr << "invalid value to -size argument: " << argv[i] << endl;
					return -1;
				}
			} else {
				cerr << "invalid value to -size argument: " << argv[i] << endl;
				return -1;
			}
		} else if(!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
			cout << "usage: " << argv[0] << " [-size <width>x<height> (optional)]\n";
			return 0;
		}
	}

	if(!Init()) {
		std::cerr << "arggg\n";
		CleanUp();
		return 0;
	}

	glutMainLoop();

	CleanUp();
	return 0;
}

void InitStateVariables();	// implementation in state_variables.cpp

bool Init() {
	glutInitWindowSize(ScreenX, ScreenY);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);

	win = glutCreateWindow("test");
	glutDisplayFunc(MainLoop);
	glutIdleFunc(MainLoop);
	glutKeyboardFunc(KeyHandler);
	glutSpecialFunc(SpecialKeyHandler);
	glutSpecialUpFunc(SpecialKeyUpHandler);
	glutPassiveMotionFunc(motion_handler);

	glutSetCursor(GLUT_CURSOR_NONE);

	InitStateVariables();
	//cam.Translate(Vector3(0, 6, 0));
	//cam.UpdateViewMatrix();

	glMatrixMode(GL_PROJECTION);
	gluPerspective(45.0f, 1.33333f, 1.0f, 1000.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	Color ambient(0.2f, 0.2f, 0.2f);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, (float*)&ambient);

	//glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glFrontFace(GL_CW);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	glFogi(GL_FOG_MODE, GL_EXP);
	glFogf(GL_FOG_DENSITY, 0.015f);
	glEnable(GL_FOG);

	lvl.LoadFromFile("data/level0.lvl");

	int start_x, start_y;
	lvl.GetStartCell(&start_x, &start_y);
	cam.Translate(lvl.GetCellCenter(start_x, start_y));

	light = Light(Vector3(0, 0, 0), Color(1.0f, 1.0f, 1.0f));
	light.SetLight(0);

	timer_reset(&timer);
	
	return true;
}

float last_lava_msg = 0.0f;
const float frame_interval = 1.0f / 50.0f;
void MainLoop() {
	static float prev_t;
	float t;
	while((t = (float)timer_getmsec(&timer) / 1000.0f) - prev_t < frame_interval) usleep(100);

    ProcessKeyPresses();

	light.SetPosition(cam.GetPosition());
	light.SetLight(0);

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	lvl.Render();

	if(lvl.GetCellFromPos(cam.GetPosition())->state == CellLava && t - last_lava_msg > 10.0f) {
		glEnable(GL_BLEND);
		glBlendFunc(GL_ONE, GL_ONE);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();
	
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glOrtho(0.0, 1.0, 1.0, 0.0, 0.0, 1.0);

		glDisable(GL_DEPTH_TEST);
		
		glBegin(GL_QUADS);
		glColor4f(0.5, 0.0, 0.0, 1.0);
		glTexCoord2f(0.0f, 1.0f);
		glVertex3f(0.0, 0.0, -0.5);
		glTexCoord2f(1.0f, 1.0f);
		glVertex3f(1.0, 0.0, -0.5);
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f(1.0, 1.0, -0.5);
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(0.0, 1.0, -0.5);
		glEnd();

		glEnable(GL_DEPTH_TEST);

		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glPopMatrix();
		glDisable(GL_BLEND);
	}

	glFlush();
	glutSwapBuffers();

	prev_t = t;
}

void CleanUp() {
	glutDestroyWindow(win);
}

void KeyHandler(unsigned char key, int x, int y) {

	switch(key) {
	case 27:
		CleanUp();
		exit(0);

	case 'f':
	case 'F':
		fullscreen = !fullscreen;

		if(fullscreen) {
			glutFullScreen();
		} else {
			glutReshapeWindow(ScreenX, ScreenY);
		}
		break;

	case 'r':
	case 'R':
		running = !running;
		break;

	case 'm':
	case 'M':
		mouse_enabled = !mouse_enabled;
		break;

	default:
		break;
	}
}

enum {
	KEY_UP		= 0,
	KEY_DOWN	= 1,
	KEY_LEFT	= 2,
	KEY_RIGHT	= 3,
	KEY_PGUP	= 4,
	KEY_PGDOWN	= 5
};

int motion_keys[6];

void SpecialKeyHandler(int key, int x, int y) {
	switch(key) {
	case GLUT_KEY_UP:
		motion_keys[KEY_UP] = 1;
		break;

	case GLUT_KEY_DOWN:
		motion_keys[KEY_DOWN] = 1;
		break;

	case GLUT_KEY_LEFT:
		motion_keys[KEY_LEFT] = 1;
		break;

	case GLUT_KEY_RIGHT:
		motion_keys[KEY_RIGHT] = 1;
	}
}
	
void SpecialKeyUpHandler(int key, int x, int y) {
	switch(key) {
	case GLUT_KEY_UP:
		motion_keys[KEY_UP] = 0;
		break;

	case GLUT_KEY_DOWN:
		motion_keys[KEY_DOWN] = 0;
		break;

	case GLUT_KEY_LEFT:
		motion_keys[KEY_LEFT] = 0;
		break;

	case GLUT_KEY_RIGHT:
		motion_keys[KEY_RIGHT] = 0;
	}
}

// checks if a disk of a specified radius is within the walls (used by RestrainMovement())
bool IsDiskInside(const Vector3 &center, float radius) {
	float step = QuarterPi / 2.0f;
	for(float a=0; a<TwoPi; a+=step) {
		Vector3 r(radius, 0, 0);
		r.Rotate(0, a, 0);
		if(lvl.GetCellFromPos(center + r)->state == CellEmpty) return false;
	}
	return true;
}

// main collision detection function for wall-player collisions
void RestrainMovement(Vector3 *fwd) {
	static const float step = QuarterPi / 4.0f;

	for(float a=0; a<Pi; a+=step) {
		Vector3 pos_dir = *fwd;
		pos_dir.Rotate(0, a, 0);
		if(IsDiskInside(cam.GetPosition() + pos_dir, 2.0f)) {
			*fwd = pos_dir;
			return;
		}

		Vector3 neg_dir = *fwd;
		neg_dir.Rotate(0, -a, 0);
		if(IsDiskInside(cam.GetPosition() + neg_dir, 2.0f)) {
			*fwd = neg_dir;
			return;
		}
	}

	*fwd = Vector3(0, 0, 0);
}

void motion_handler(int x, int y) {
	static int prev_x = -1, prev_y;
	static float angle, pitch;

	if(mouse_enabled && prev_x != -1) {
		angle += -0.005 * ((float)x - (float)prev_x);
		pitch += 0.005 * ((float)y - (float)prev_y);

		if(pitch < -HalfPi) pitch = 0.0;
		if(pitch > HalfPi) pitch = HalfPi;

		cam.ResetRotation();
		cam.Rotate(Vector3(0, 1, 0), angle);
		cam.Rotate(Vector3(1, 0, 0), pitch);
	}
	prev_x = x;
	prev_y = y;
}

// processes the key presses
void ProcessKeyPresses() {

	const float rot_angle = running ? 0.05f : 0.04f;
	
	if(motion_keys[KEY_LEFT]) {
        cam.Rotate(Vector3(0, 1, 0), rot_angle);
	}

	if(motion_keys[KEY_RIGHT]) {
		cam.Rotate(Vector3(0, 1, 0), -rot_angle);
	}

	if(motion_keys[KEY_PGUP]) {
        cam.Rotate(Vector3(1,0,0), rot_angle);
	}

	if(motion_keys[KEY_PGDOWN]) {
		cam.Rotate(Vector3(1,0,0), -rot_angle);
	}

	Vector3 fwd(0, 0, running ? -1.1 : -0.62);
	fwd.Transform(cam.GetRotation());

	if(motion_keys[KEY_UP]) {
		RestrainMovement(&fwd);
		cam.Translate(fwd);
	}

	if(motion_keys[KEY_DOWN]) {
		fwd = -fwd;
        RestrainMovement(&fwd);
		cam.Translate(fwd);
	}

	cam.UpdateViewMatrix();
}
