#include <iostream>
#include <fstream>
#include <GL/gl.h>
#include "terrain.h"

using std::cerr;

Level::Level() {

	for(int i=0; i<LEVEL_SIZE; i++) {
		for(int j=0; j<LEVEL_SIZE; j++) {
			cells[i][j].state = CellEmpty;
		}
	}
}

struct Quad {
	Vertex verts[5];
	Triangle tris[4];
};

static float HalfCellSize = (float)(CELL_SIZE>>1);

//////// a function that generates the geometry for the floor /////////
void CreateFloor(int x, int y, Quad *quad) {
	quad->verts[0] = Vertex(Vector3(-HalfCellSize, 0, -HalfCellSize), 1.0f, 0.0f, 0.5f);
	quad->verts[1] = Vertex(Vector3(HalfCellSize, 0, -HalfCellSize), 1.0f, 1.0f, 0.5f);
	quad->verts[2] = Vertex(Vector3(HalfCellSize, 0, HalfCellSize), 1.0f, 1.0f, 0.0f);
	quad->verts[3] = Vertex(Vector3(-HalfCellSize, 0, HalfCellSize), 1.0f, 0.0f, 0.0f);
	quad->verts[4] = Vertex(Vector3(0, 0, 0), 1.0f, 0.5f, 0.25f);
	quad->tris[0] = Triangle(0, 1, 4);
	quad->tris[1] = Triangle(1, 2, 4);
	quad->tris[2] = Triangle(2, 3, 4);
	quad->tris[3] = Triangle(3, 0, 4);

	
	Vector3 offset((float)((x - LEVEL_SIZE/2) * CELL_SIZE), 0, (float)((y - LEVEL_SIZE/2) * CELL_SIZE));
	for(int i=0; i<5; i++) {
		quad->verts[i].pos += offset;
	}
}

void CreatePit(int x, int y, Quad *quad) {
	CreateFloor(x, y, quad);

	quad->verts[4].pos.y = -100.0f;
	quad->verts[4].u = 30.0f;
	quad->verts[4].v = 30.0f;
	quad->verts[4].color = Color(1.0f, 0.0f, 0.0f);
}


//////// a function that generates the geometry for the ceiling /////////
void CreateCeiling(int x, int y, Quad *quad) {

	CreateFloor(x, y, quad);
	
	Vector3 offset(0, CELL_SIZE, 0);
	for(int i=0; i<5; i++) {
		quad->verts[i].pos += offset;
		quad->verts[i].v *= 2.0f;
	}

	// change the winding
	for(int i=0; i<4; i++) {
		Index temp = quad->tris[i].verts[2];
		quad->tris[i].verts[2] = quad->tris[i].verts[1];
		quad->tris[i].verts[1] = temp;
	}
}

#define WALL_NORTH	0
#define WALL_SOUTH	1
#define WALL_EAST	2
#define WALL_WEST	3

void CreateWall(int x, int y, int dir, Quad *quad) {

	quad->verts[0] = Vertex(Vector3(-HalfCellSize, CELL_SIZE, -HalfCellSize), 1.0f, 0.0f, 1.0f);
	quad->verts[1] = Vertex(Vector3(HalfCellSize, CELL_SIZE, -HalfCellSize), 1.0f, 1.0f, 1.0f);
	quad->verts[2] = Vertex(Vector3(HalfCellSize, 0, -HalfCellSize), 1.0f, 1.0f, 0.5f);
	quad->verts[3] = Vertex(Vector3(-HalfCellSize, 0, -HalfCellSize), 1.0f, 0.0f, 0.5f);
	quad->verts[4] = Vertex(Vector3(0, HalfCellSize, -HalfCellSize), 1.0f, 0.5f, 0.75f);
	quad->tris[0] = Triangle(0, 1, 4);
	quad->tris[1] = Triangle(1, 2, 4);
	quad->tris[2] = Triangle(2, 3, 4);
	quad->tris[3] = Triangle(3, 0, 4);

	Vector3 offset((float)((x - LEVEL_SIZE/2) * CELL_SIZE), 0, (float)((y - LEVEL_SIZE/2) * CELL_SIZE));
	if(dir == WALL_SOUTH) {
		Matrix4x4 mat;
		mat.Rotate(0, Pi, 0);
		mat.Translate(offset.x, offset.y, offset.z);

		for(int i=0; i<5; i++) {
			quad->verts[i].pos.Transform(mat);
		}
	} else if(dir == WALL_EAST) {
		Matrix4x4 mat;
		mat.Rotate(0, HalfPi, 0);
		mat.Translate(offset.x, offset.y, offset.z);

		for(int i=0; i<5; i++) {
			quad->verts[i].pos.Transform(mat);
		}
	} else if(dir == WALL_WEST) {
		Matrix4x4 mat;
		mat.Rotate(0, -HalfPi, 0);
		mat.Translate(offset.x, offset.y, offset.z);

		for(int i=0; i<5; i++) {
			quad->verts[i].pos.Transform(mat);
		}
	} else {	// NORTH or undefined
		for(int i=0; i<5; i++) {
			quad->verts[i].pos += offset;
		}
	}
}

void AddToLists(Quad *q, std::list<Vertex> *vlist, std::list<Triangle> *tlist) {

	Index index_offset = (Index)vlist->size();
	for(int i=0; i<5; i++) {
		vlist->push_back(q->verts[i]);
	}

	for(int i=0; i<4; i++) {
		for(int j=0; j<3; j++) {
			q->tris[i].verts[j] += index_offset;
		}
		tlist->push_back(q->tris[i]);
	}
}

/* ---- GenerateGeometry [private function] ----
** Generates the geometry (mesh data) of the maze
** based on cell states.
** Only walls, floors and ceilings are generated, not
** entities
*/
void Level::GenerateGeometry() {

	std::list<Vertex> maze_vert_list;
	std::list<Triangle> maze_tri_list;

	std::list<Vertex> ceil_vert_list;
	std::list<Triangle> ceil_tri_list;

	std::list<Vertex> lava_vert_list;
	std::list<Triangle> lava_tri_list;

	Quad quad;

	for(int i=0; i<LEVEL_SIZE; i++) {
		for(int j=0; j<LEVEL_SIZE; j++) {
			if(cells[i][j].state == CellEmpty) continue;	// no geometry on empty cells

			// create the ceiling
			CreateCeiling(i, j, &quad);
			AddToLists(&quad, &ceil_vert_list, &ceil_tri_list);

			if(cells[i][j].state != CellLava) {
				// if the cell is *not* a lava pit, create a floor
				CreateFloor(i, j, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			} else {
				CreatePit(i, j, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			}

			if(cells[i+1][j].state == CellEmpty) {
				CreateWall(i, j, WALL_WEST, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			}

			if(cells[i-1][j].state == CellEmpty) {
				CreateWall(i, j, WALL_EAST, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			}

			if(cells[i][j+1].state == CellEmpty) {
				CreateWall(i, j, WALL_SOUTH, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			}

			if(cells[i][j-1].state == CellEmpty) {
				CreateWall(i, j, WALL_NORTH, &quad);
				AddToLists(&quad, &maze_vert_list, &maze_tri_list);
			}
		}
	}

	////////// fill the data arrays with the maze geometry /////////
	Vertex *varray = new Vertex[maze_vert_list.size()];
	Triangle *tarray = new Triangle[maze_tri_list.size()];

	std::list<Vertex>::iterator viter = maze_vert_list.begin();
	for(unsigned int i=0; i<maze_vert_list.size(); i++) {
		varray[i] = *viter++;
	}

	std::list<Triangle>::iterator titer = maze_tri_list.begin();
	for(unsigned int i=0; i<maze_tri_list.size(); i++) {
		tarray[i] = *titer++;
	}

	CalculateTriangleNormals(tarray, (unsigned long)maze_tri_list.size(), varray);
	CalculateVertexNormals(varray, (unsigned long)maze_vert_list.size(), tarray, (unsigned long)maze_tri_list.size());

	maze.SetGeometry(varray, tarray, (unsigned long)maze_vert_list.size(), (unsigned long)maze_tri_list.size());

	delete [] varray;
	delete [] tarray;

	////////// fill the data arrays with the ceiling geometry /////////

	varray = new Vertex[ceil_vert_list.size()];
	tarray = new Triangle[ceil_tri_list.size()];

	viter = ceil_vert_list.begin();
	for(unsigned int i=0; i<ceil_vert_list.size(); i++) {
		varray[i] = *viter++;
	}

	titer = ceil_tri_list.begin();
	for(unsigned int i=0; i<ceil_tri_list.size(); i++) {
		tarray[i] = *titer++;
	}

	CalculateTriangleNormals(tarray, (unsigned long)ceil_tri_list.size(), varray);
	CalculateVertexNormals(varray, (unsigned long)ceil_vert_list.size(), tarray, (unsigned long)ceil_tri_list.size());

	ceiling.SetGeometry(varray, tarray, (unsigned long)ceil_vert_list.size(), (unsigned long)ceil_tri_list.size());

	delete [] varray;
	delete [] tarray;

	////////////////////////// create the lava under the dungeon //////////////

	varray = new Vertex[4];
	tarray = new Triangle[2];

	float HalfLevelSize = (float)(CELL_SIZE * LEVEL_SIZE) / 2.0f;
	const float lava_level = -2.0f;

	varray[0] = Vertex(Vector3(-HalfLevelSize, lava_level, -HalfLevelSize), 1.0f, 0.0f, 20.0f);
	varray[1] = Vertex(Vector3(HalfLevelSize, lava_level, -HalfLevelSize), 1.0f, 20.0f, 20.0f);
	varray[2] = Vertex(Vector3(HalfLevelSize, lava_level, HalfLevelSize), 1.0f, 20.0f, 0.0f);
	varray[3] = Vertex(Vector3(-HalfLevelSize, lava_level, HalfLevelSize), 1.0f, 0.0f, 0.0f);
	tarray[0] = Triangle(0, 1, 2);
	tarray[1] = Triangle(0, 2, 3);

	CalculateTriangleNormals(tarray, 2, varray);
	CalculateVertexNormals(varray, 4, tarray, 2);

	lava.SetGeometry(varray, tarray, 4, 2);
		
	delete [] varray;
	delete [] tarray;

	Material material;

	material.texmap = LoadTexture("data/wall2.png");
	maze.SetMaterial(material);

	material.texmap = LoadTexture("data/roof.png");
	ceiling.SetMaterial(material);

	material.emissive = Color(1.0f, 1.0f, 1.0f);
	material.texmap = LoadTexture("data/lava.png");
	lava.SetMaterial(material);
}

// loads the level description from a text file (see level0.lvl)
void Level::LoadFromFile(const char *fname) {

	std::ifstream file(fname);
	if(!file.is_open()) return;

    char line[LEVEL_SIZE+1];
	file.getline(line, LEVEL_SIZE);
	if(!strstr(line, "GAME_LEVEL")) {
		cerr << "Error: invalid level file\n";
		return;
	}
	
	// if the file is a valid game level then
	for(int i=0; i<LEVEL_SIZE; i++) {
		file.getline(line, LEVEL_SIZE+1);
		if(strlen(line) != LEVEL_SIZE) {
			cerr << "Error: invalid level file\n";
			return;
		}

		char *cell = line;
		for(int j=0; j<LEVEL_SIZE; j++) {
			switch(*cell++) {
			case '#':
				cells[j][i].state = CellEmpty;
				break;

			case 'S':
				cells[j][i].state = CellStart;
				break;

			case 'L':
				cells[j][i].state = CellLava;
				break;

			case ' ':
			case 'M':
			case 'm':
				cells[j][i].state = CellCorridor;
				break;

			default:
				cerr << "Error: Unrecognized symbol in level file, leaving cell empty";
			}
		}
	}

	GenerateGeometry();

}

void Level::GetStartCell(int *x, int *y) {

	for(int ly=0; ly<LEVEL_SIZE; ly++) {
		for(int lx=0; lx<LEVEL_SIZE; lx++) {
			if(cells[lx][ly].state == CellStart) {
				*x = lx;
				*y = ly;
				return;
			}				
		}
	}

					
}

LevelCell *Level::GetCellFromPos(const Vector3 &pos) {
	int x = (int)((pos.x - (float)CELL_SIZE / 2.0f) / (float)CELL_SIZE + (float)LEVEL_SIZE / 2.0f) + 1;
	int y = (int)((pos.z - (float)CELL_SIZE / 2.0f) / (float)CELL_SIZE + (float)LEVEL_SIZE / 2.0f) + 1;

	return &cells[x][y];
}

Vector3 Level::GetCellCenter(int x, int y) {
	Vector3 pos;
	pos.x = (x - (LEVEL_SIZE >> 1)) * CELL_SIZE;
	pos.y = 6.0f;
	pos.z = (y - (LEVEL_SIZE >> 1)) * CELL_SIZE;

	return pos;
}

void Level::Render() {

	maze.Render();
	ceiling.Render();

	glDisable(GL_FOG);
	lava.Render();
	glEnable(GL_FOG);
}
